import os
import webbrowser
from pathlib import Path

from pkg_resources import resource_filename

from .. import download_data
from ... import Flow
from ...importer import ImportExtensions
from ...logging import default_logger


def hello_world(args):
    Path(args.workdir).mkdir(parents=True, exist_ok=True)

    with ImportExtensions(required=True, help_text='this demo requires Pytorch and Transformers to be installed, '
                                                   'if you haven\'t, please do `pip install jina[torch,transformers]`'):
        import transformers, torch
        assert [torch, transformers]  #: prevent pycharm auto remove the above line

    targets = {
        'people-img': {
            'url': args.index_data_url,
            'filename': os.path.join(args.workdir, 'dataset.zip')
        }
    }

    # download the data
    download_data(targets, args.download_proxy, task_name='download zip data')
    import zipfile
    with zipfile.ZipFile(targets['people-img']['filename'], 'r') as fp:
        fp.extractall(args.workdir)

    # this envs are referred in index and query flow YAMLs
    os.environ['HW_WORKDIR'] = args.workdir
    os.environ['PATH'] += os.pathsep + os.path.join(resource_filename('jina', 'resources'), 'multimodal')

    # now comes the real work
    # load index flow from a YAML file

    # index it!
    f = Flow.load_config('flow-index.yml')
    with f, open(f'{args.workdir}/people-img/meta.csv') as fp:
        f.index_csv(fp)

    # search it!

    f = Flow.load_config('flow-query.yml')
    # switch to REST gateway
    f.use_rest_gateway(args.port_expose)

    with f:
        try:
            webbrowser.open(args.demo_url, new=2)
        except:
            pass  # intentional pass, browser support isn't cross-platform
        finally:
            default_logger.success(f'You should see a demo page opened in your browser, '
                                   f'if not, you may open {args.demo_url} manually')
        if not args.unblock_query_flow:
            f.block()



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import warnings
from typing import Optional

from . import BaseExecutableDriver, FastRecursiveMixin, RecursiveMixin
from ..types.sets import DocumentSet


class BaseEncodeDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`encode` by default """

    def __init__(self,
                 executor: str = None,
                 method: str = 'encode',
                 *args, **kwargs):
        super().__init__(executor, method, *args, **kwargs)


class EncodeDriver(FastRecursiveMixin, BaseEncodeDriver):
    """Extract the content from documents and call executor and do encoding
    """

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        contents, docs_pts = docs.all_contents

        if docs_pts:
            embeds = self.exec_fn(contents)
            if len(docs_pts) != embeds.shape[0]:
                self.logger.error(
                    f'mismatched {len(docs_pts)} docs from level {docs_pts[0].granularity} '
                    f'and a {embeds.shape} shape embedding, the first dimension must be the same')
            for doc, embedding in zip(docs_pts, embeds):
                doc.embedding = embedding


class LegacyEncodeDriver(RecursiveMixin, BaseEncodeDriver):
    """Extract the content from documents and call executor and do encoding

    .. note::

        ``batch_size`` is specially useful when the same EncoderExecutor can be used for documents of different granularities
         (chunks, chunks of chunks ...)

    .. warning::

        ``batch_size`` parameter was added to cover the case where root documents had very few chunks, and the encoder executor could
        then only process them in batches of the chunk size of each document, which did not lead to the full use of batching capabilities
        of the powerful Executors

    :param batch_size: number of documents to be used simultaneously in the encoder :meth:_apply_all.
    :param *args: *args for super
    :param **kwargs: **kwargs for super
    """

    class CacheDocumentSet:
        """Helper class to accumulate documents from different DocumentSets in a single DocumentSet
         to help guarantee that the encoder driver can consume documents in fixed batch sizes to allow
         the EncoderExecutors to leverage its batching abilities.
         It is useful to have batching even when chunks are involved"""

        def __init__(self,
                     capacity: Optional[int] = None,
                     *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.capacity = capacity
            self._doc_set = DocumentSet(docs_proto=[])

        @property
        def available_capacity(self):
            """The capacity left in the cache


            .. # noqa: DAR201
            """
            return self.capacity - len(self._doc_set)

        def cache(self, docs: DocumentSet):
            """Cache the docs in DocumentSet.

            :param docs: the DocumentSet to cache
            :return: the subset of the docs
            """
            docs_to_append = min(len(docs), self.available_capacity)
            self._doc_set.extend(docs[: docs_to_append])
            return DocumentSet(docs[docs_to_append:])

        def __len__(self):
            return len(self._doc_set)

        def get(self):
            """Get the DocumentSet


            .. # noqa: DAR201
            """
            return self._doc_set

    def __init__(self,
                 batch_size: Optional[int] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(f'this drivers will be removed soon, use {EncodeDriver!r} instead', DeprecationWarning)
        self.batch_size = batch_size
        if self.batch_size:
            self.cache_set = LegacyEncodeDriver.CacheDocumentSet(capacity=self.batch_size)
        else:
            self.cache_set = None

    def __call__(self, *args, **kwargs):
        """Traverse the documents with the Driver.

        :param *args: *args for ``_traverse_apply``
        :param **kwargs: **kwargs for ``_traverse_apply``
        """
        self._traverse_apply(self.docs, *args, **kwargs)
        self._empty_cache()

    def _apply_batch(self, batch: 'DocumentSet'):
        contents, docs_pts = batch.all_contents

        if docs_pts:
            embeds = self.exec_fn(contents)
            if embeds is None:
                self.logger.error(
                    f'{self.exec_fn!r} returns nothing, you may want to check the implementation of {self.exec!r}')
            elif len(docs_pts) != embeds.shape[0]:
                self.logger.error(
                    f'mismatched {len(docs_pts)} docs from level {docs_pts[0].granularity} '
                    f'and a {embeds.shape} shape embedding, the first dimension must be the same')
            else:
                for doc, embedding in zip(docs_pts, embeds):
                    doc.embedding = embedding

    def _empty_cache(self):
        if self.batch_size:
            cached_docs = self.cache_set.get()
            if len(cached_docs) > 0:
                self._apply_batch(cached_docs)
            self.cache_set = LegacyEncodeDriver.CacheDocumentSet(capacity=self.batch_size)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        if self.cache_set is not None:
            left_docs = self.cache_set.cache(docs)
            while len(left_docs) > 0:
                self._empty_cache()
                left_docs = self.cache_set.cache(left_docs)
            if self.cache_set.available_capacity == 0:
                self._empty_cache()
        else:
            self._apply_batch(docs)



"""Module containing the base parser for arguments of Jina."""
import argparse

from .helper import _chf


def set_base_parser():
    """Set the base parser

    :return: the parser
    """
    from .. import __version__
    from ..helper import colored, get_full_version, format_full_version_info
    # create the top-level parser
    urls = {
        'Jina 101': ('🐣', 'https://101.jina.ai'),
        'Docs': ('📚', 'https://docs.jina.ai'),
        'Examples': ('🚀‍', 'https://learn.jina.ai'),
        'Code': ('🧑‍💻', 'https://opensource.jina.ai'),
        'Hiring!': ('🙌', 'https://career.jina.ai')
    }
    url_str = '\n'.join(f'- {v[0]:<10} {k:10.10}\t{colored(v[1], "cyan", attrs=["underline"])}' for k, v in urls.items())

    parser = argparse.ArgumentParser(
        epilog=f'''
Jina (v{colored(__version__, "green")}) is the cloud-native neural search solution powered by AI & deep learning.
It is a universal solution to large-scale index and query of unstructured & multimedia data.

{url_str}

''',
        formatter_class=_chf,
        description='Command Line Interface of `%(prog)s`'
    )
    parser.add_argument('-v', '--version', action='version', version=__version__,
                        help='Show Jina version')

    parser.add_argument('-vf', '--version-full', action='version',
                        version=format_full_version_info(*get_full_version()),
                        help='Show Jina and all dependencies\' versions')
    return parser



import asyncio
import json
import uuid
from pathlib import Path

from fastapi import APIRouter, WebSocket
from fastapi.exceptions import HTTPException
from fastapi.responses import FileResponse
from starlette.endpoints import WebSocketEndpoint
from starlette.types import Receive, Scope, Send

from ... import daemon_logger, jinad_args
from ...stores.helper import get_workspace_path

router = APIRouter(tags=['logs'])


@router.get(
    path='/logs/{workspace_id}/{log_id}'
)
async def _export_logs(
        workspace_id: uuid.UUID,
        log_id: uuid.UUID
):
    filepath = get_workspace_path(workspace_id, log_id, 'logging.log')
    if not Path(filepath).is_file():
        raise HTTPException(status_code=404, detail=f'log file {log_id} not found in workspace {workspace_id}')
    else:
        return FileResponse(filepath)


class LogStreamingEndpoint(WebSocketEndpoint):

    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        super().__init__(scope, receive, send)
        # Accessing path / query params from scope in ASGI
        # https://asgi.readthedocs.io/en/latest/specs/www.html#websocket-connection-scope
        info = self.scope.get('path').split('/')
        workspace_id = info[-2]
        log_id = info[-1]
        self.filepath = get_workspace_path(workspace_id, log_id, 'logging.log')
        self.active_clients = []

    async def on_connect(self, websocket: WebSocket) -> None:
        await websocket.accept()

        self.client_details = f'{websocket.client.host}:{websocket.client.port}'
        self.active_clients.append(websocket)
        daemon_logger.info(f'{self.client_details} is connected to stream logs!')

        if jinad_args.no_fluentd:
            daemon_logger.warning(f'{self.client_details} asks for logstreaming but fluentd is not available')
            return

        # on connection the fluentd file may not flushed (aka exist) yet
        while not Path(self.filepath).is_file():
            daemon_logger.debug(f'still waiting {self.filepath} to be ready...')
            await asyncio.sleep(1)

        with open(self.filepath) as fp:
            fp.seek(0, 2)
            daemon_logger.success(f'{self.filepath} is ready for streaming')
            while websocket in self.active_clients:
                line = fp.readline()  # also possible to read an empty line
                if line:
                    payload = None
                    try:
                        payload = json.loads(line)
                    except json.decoder.JSONDecodeError:
                        daemon_logger.warning(f'JSON decode error on {line}')

                    if payload:
                        from websockets import ConnectionClosedOK
                        try:
                            await websocket.send_json(payload)
                        except ConnectionClosedOK:
                            break
                else:
                    await asyncio.sleep(0.1)

    async def on_disconnect(self, websocket: WebSocket, close_code: int) -> None:
        self.active_clients.remove(websocket)
        daemon_logger.info(f'{self.client_details} is disconnected')


# TODO: adding websocket in this way do not generate any docs
#  see: https://github.com/tiangolo/fastapi/issues/1983
router.add_websocket_route(path='/logstream/{workspace_id}/{log_id}',
                           endpoint=LogStreamingEndpoint)



import os

from jina import Segmenter, Crafter


class SimpleCrafter(Crafter):

    def craft(self, tags):
        return {'text': tags['caption'],
                'uri': f'{os.environ["HW_WORKDIR"]}/people-img/{tags["image"]}'}


class BiSegmenter(Segmenter):

    def segment(self, text, uri):
        return [
            {'text': text, 'mime_type': 'text/plain'},
            {'uri': uri, 'mime_type': 'image/jpeg'}
        ]



from jina.executors.evaluators.text import BaseTextEvaluator


class HammingDistanceEvaluator(BaseTextEvaluator):
    """A:class:`HammingDistanceEvaluator` Gives the Hamming distance between result and groundtruth string..
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def evaluate(self, actual: str, desired: str):
        if len(actual) != len(desired):
            raise ValueError('Undefined for sequences of unequal length.')
        dist_counter = 0
        for n in range(len(actual)):
            if actual[n] != desired[n]:
                dist_counter += 1
        return dist_counter



from .custom import build_pydantic_model

FlowModel = build_pydantic_model(model_name='FlowModel', module='flow')
PodModel = build_pydantic_model(model_name='PodModel', module='pod')
PeaModel = build_pydantic_model(model_name='PeaModel', module='pea')